/*
 * Copyright 2023 Arne Vajh�j.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

package dk.vajhoej.record;

import java.lang.reflect.Array;
import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;

/**
 * Class MapHelper contains helper methods to convert between Map and atruct objects.
 *
 */
public class MapHelper {
    private static Logger log = Logger.getLogger(MapHelper.class.getName());
    private static void stuffList(Object o, List<Object> list) throws IllegalArgumentException, IllegalAccessException, RecordException, InstantiationException {
       	for(int i = 0; i < list.size(); i++) {
     		if(o.getClass().getComponentType().isArray() && list.get(i) instanceof List) {
    			stuffList(Array.get(o, i), (List<Object>)list.get(i));
    		} else if(o.getClass().getComponentType().getAnnotation(Struct.class) != null && list.get(i) instanceof Map) {
    			Object o2 = o.getClass().getComponentType().newInstance();
    			stuffMap(o2, (Map<String,Object>)list.get(i));
    			Array.set(o, i, o2);
    		} else {
    			Array.set(o, i, list.get(i));
    		}
    	}
       	for(int i = list.size(); i < Array.getLength(o); i++) {
       		Array.set(o, i, o.getClass().getComponentType().newInstance());
       	}
    }
    private static void stuffMap(Object o, Map<String,Object> map) throws IllegalArgumentException, IllegalAccessException, RecordException, InstantiationException {
    	StructInfo si = StructInfoCache.analyze(o.getClass());
		for(FieldInfo fi : si.getFields()) {
			Field f = fi.getField();
			Object val = map.get(f.getName());
			if(map.containsKey(f.getName())) {
				if(f.getType().isArray() && val instanceof List) {
					stuffList(f.get(o), (List<Object>)val);
				} else if(fi.getStructType() == FieldType.STRUCT && val instanceof Map) {
					stuffMap(f.get(o), (Map<String,Object>)val);
				} else {
					f.set(o, val);
				}
			}
		}
    }
	/**
	 * Convert from Map to a struct object.
	 * @param <T> struct type
	 * @param t struct class
	 * @param map input map
	 * @return output struct object
	 * @throws RecordException if unable to convert (like due to reflection issue)
	 */
	public static <T> T fromMap(Class<T> t, Map<String,Object> map) throws RecordException {
    	try {
			T o = t.newInstance();
			stuffMap(o, map);
			log.finest(map + " converted to " + o);
			return o;
		} catch (InstantiationException e) {
			log.fine("fromMap threw: " + e);
			throw new RecordException("Error converting from Map to Object", e);
		} catch (IllegalAccessException e) {
			log.fine("fromMap threw: " + e);
			throw new RecordException("Error converting from Map to Object", e);
		} catch (IllegalArgumentException e) {
			log.fine("fromMap threw: " + e);
			throw new RecordException("Error converting from Map to Object", e);
		}
 		
	}
	private static List<Object> extractList(Object o) throws IllegalArgumentException, IllegalAccessException, RecordException {
		List<Object> lst = new ArrayList<Object>();
		for(int i = 0; i < Array.getLength(o); i++) {
			Object elm = Array.get(o, i);
			if(elm.getClass().isArray()) {
				lst.add(extractList(elm));
			} else if(elm instanceof Map) {
				lst.add(extractMap(elm));
			} else {
				lst.add(elm);
			}
		}
		return lst;
	}
	private static Map<String,Object> extractMap(Object o) throws RecordException, IllegalArgumentException, IllegalAccessException {
		Map<String,Object> map = new HashMap<String,Object>();
		StructInfo si = StructInfoCache.analyze(o.getClass());
		for(FieldInfo fi : si.getFields()) {
			Field f = fi.getField();
			if(f.getAnnotation(ArrayField.class) != null) {
				map.put(f.getName(), extractList(f.get(o)));
			} else if(fi.getStructType() == FieldType.STRUCT) {
				map.put(f.getName(), toMap(f.get(o)));
			} else {
				map.put(f.getName(), f.get(o));
			}
		}
		return map;
	}
	/**
	 * Convert frm struct object to Map.
	 * @param <T> struct type
	 * @param o input struct object
	 * @return output Map
	 * @throws RecordException if unable to convert (like due to reflection issue)
	 */
	public static <T> Map<String,Object> toMap(T o) throws RecordException {
		try {
			Map<String,Object> map = extractMap(o);
			log.finest(o + " converted to " + map);
			return map;
		} catch (IllegalArgumentException e) {
			log.fine("toMap threw: " + e);
			throw new RecordException("Error converting from Object to Map", e);
		} catch (IllegalAccessException e) {
			log.fine("toMap threw: " + e);
			throw new RecordException("Error converting from Object to Map", e);
		}
	}
}
